;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Game of Go
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; vim: set syntax=asm6809
; assemble with asm6809: https://www.6809.org.uk/asm6809/
; git clone http://www.6809.org.uk/git/asm6809.git
; sudo apt-get install automake bison flex
; ./autogen.sh
; ./configure
; make && sudo make install

width		equ	9
height		equ	9

empty		equ	0
black		equ	8
white		equ	16
outside		equ	24

bg_color	equ	$0f
cursor_color	equ	$01
screen_location	equ	19 - width / 2 + 40 * (100 - 4 * (height + 2))

	org	$6080
	setdp	#direct_page >> 8

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Display board
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
display_board
	ldx	#board
	ldu	#screen_location

	lda	#width + 1
	sta	loop_index_1
display_loop_over_width

	lda	#height + 1
	sta	loop_index_2
display_loop_over_height
	cmpx	cursor
	bne	display_loop_not_cursor
	stu	cursor_location
display_loop_not_cursor
	lda	, x+
	bsr	write_block ; modified later to print_character
display_bsr

	dec	loop_index_2
	bge	display_loop_over_height

	leau	(1 - 320 * (height + 2)), u
	dec	loop_index_1
	bge	display_loop_over_width
	rts

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; print character (A, U)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
print_character
	ldy	#char_table
	leay	a, y

	ldb	#7
print_character_loop
	lda	b, y
	sta	, u
	leau	40, u
	decb
	bge	print_character_loop

	rts

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Write cursor at U
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
write_cursor
	ldu	cursor_location
write_block
	lda	#7
write_cursor_loop
	stb	, u
	leau	40, u
	deca
	bge	write_cursor_loop

	rts

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Main
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
main
	export	main
	lda	#direct_page >> 8
	tfr	a, dp

;
; Background color
;
	lda	$A7C0
	anda	#$FE
	sta	$A7C0
	ldb	#bg_color
	bsr	display_board
	lda	#print_character - display_bsr
	sta	display_bsr - 1
	bsr	process_key_Home

;
; Main loop: scan keyboard, process keys, display updated cursor and board
;
main_loop

;
; Board display
;
	inc	$A7C0
	bsr	display_board
	lda	to_move
	bsr	print_character
	ldx	#$A7C0
	dec	, x+

;
; Cursor
;
	ldb	#cursor_color
	bsr	write_cursor

;
; Wait for key up
;
	lda	key
wait_for_key_up_loop
	sta	, x
	lda	, x
	bpl	wait_for_key_up_loop

;
; Wait for some time?
;

;
; Wait for next key
;
	ldu	#keys
wait_for_key
	ldb	#number_of_keys - 1
wait_for_key_next
	lda	b, u
	sta	, x
	lda	, x
	bpl	wait_for_key_end
	decb
	bge	wait_for_key_next
	bra	wait_for_key
wait_for_key_end
	sta	key

;
; Process key and loop
;
	ldu	#key_processor
	lda	b, u
	sta	key_bsr - 1
	bsr	key_bsr
key_bsr

	bra	main_loop

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Move cursor with arrows
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
process_arrow
	ldx	#direction
	ldb	b, x
	ldx	cursor
	leax	b, x
	ldb	, x
	cmpb	#outside
	beq	process_arrow_rts
	stx	cursor
	ldb	#bg_color
	bsr	write_cursor
process_arrow_rts
	rts

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Navigate game history
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
process_key_Home
	ldx	#history
	bra	init_before_replaying
process_key_R
	ldx	game_index
	cmpx	game_end
	beq	make_move_rts
	leax	+2, x
	bra	ready_to_move_forward
process_key_E
	ldx	game_end
ready_to_move_forward
	ldu	game_index
replay_game
	stx	game_index
replay_game_loop
	cmpu	game_index
	beq	make_move_rts
	ldx	, u++
	pshs	u
	bsr	make_move
	puls	u
	bra	replay_game_loop
process_key_U
	ldx	game_index
	cmpx	#history
	beq	make_move_rts
	leax	-2, x

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Initialize empty board (preserve x)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
init_before_replaying
	ldd	#((black << 8) | white)
	std	to_move
	clr	ko	; Tricky: clearing one byte is enough

	lda	#outside
	ldu	#board
init_outside_loop
	sta	, u+
	cmpu	#board + (width + 2) * (height + 2)
	bne	init_outside_loop

	ldu	#board + 1 + 1 * (height + 2)
	lda	#width - 1
init_width_loop
	ldb	#height - 1
init_height_loop
	clr	, u+
	decb
	bge	init_height_loop
	leau	2, u
	deca
	bge	init_width_loop

	ldu	#history
	bra	replay_game

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Play move at cursor or pass
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
process_key_space
	ldx	cursor
	bra	test_length_and_make_move
process_key_P
	ldx	zeroes
test_length_and_make_move
	ldu	game_index
	cmpu	#history + max_game_length * 2
	bge	make_move_rts
	bsr	make_move
	bcs	make_move_rts
	ldu	game_index
	ldx	move_address
	stx	, u++
	stu	game_index
	stu	game_end
make_move_rts
	rts

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Make move (in X)
; On return, C condition indicates illegal move
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
make_move
	stx	move_address

;
; Pass
;
	bne	not_pass
	clr	ko
	bra	change_side_to_move
not_pass

;
; Check that the point is empty
;
	tst	, x
	bne	return_illegal_move

;
; Test for ko
;
	cmpx	ko
	beq	return_illegal_move

;
; Put stone
;
	lda	to_move
	sta	, x

;
; Captures
;
	ldu	#stack_bottom
	ldy	#relative_dir
	clr	capture_size
	lda	#3
	sta	loop_index_1
make_move_capture_dir_loop
	lda	loop_index_1
	ldb	a, y
	leax	b, x
	lda	, x
	cmpa	opponent
	bne	capture_next_direction
	clra
	deca
	sta	liberties
	bsr	count_liberties

capture_next_direction
	dec	loop_index_1
	bge	make_move_capture_dir_loop

;
; Check that the string has liberties
;
	ldx	move_address
	clr	liberties
	bsr	count_liberties
	lda	liberties
	bne	new_string_has_liberties
	clr	, x
return_illegal_move
	coma
	rts
new_string_has_liberties

;
; Test for ko
;
	adda	string_size ; Note: liberties + string_size < 256
	ldb	capture_size
	cmpd	#$0201
	bne	no_ko
	ldx	-4, u ; liberty of ko is second in stack
no_ko
	stx	ko

;
; Change side to move
;
change_side_to_move
	lda	opponent
	ldb	to_move
	std	to_move

	clra
	rts

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Count liberties, and string size of string at X.
; Y is relative_dir
; Preserve X and Y
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
count_liberties
	sts	system_bottom
	clr	string_size
	ldb	, x
	stb	string_stone
	clra
	bra	count_liberties_start_loop
count_liberties_loop
	puls	x
	inc	string_size
	lda	#3
count_liberties_dir_loop
	ldb	a, y
	leax	b, x
	ldb	, x
	bne	count_liberties_not_empty
	inc	liberties
	bra	count_liberties_mark
count_liberties_not_empty
	cmpb	string_stone
	bne	count_liberties_next_dir
count_liberties_start_loop
	pshs	x
count_liberties_mark
	inc	, x
	pshu	x
count_liberties_next_dir
	deca
	bge	count_liberties_dir_loop
	cmps	system_bottom
	bne	count_liberties_loop

clear_mark_loop
	pulu	x
	dec	, x
	tst	liberties
	bpl	clear_mark_no_capture
	clr	, x
	inc	capture_size
clear_mark_no_capture
	cmpu	#stack_bottom
	bne	clear_mark_loop

	rts

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Neural network
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;neural_network
;	ldb	, u+	; 6

;	orb	, s+	; 6 2/3

;	orb	, s	; 4 1/3
;	leas	9, s	; 5 1/3

;	ldb	b, y	; 5
;	abx		; 3


;	ldb	#-N*&80	; 4 (to compensate unsigned), once, at initialization

; Total for 3x3: 6 * (6 + 6 + 5 + 3) + 3 * (6 + 4 + 5 + 5 + 3) = 189
; a is free for horizontal loop

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Constants
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
number_of_keys	equ	key_processor - keys

keys
key_right	fcb	2 * $19
key_left	fcb	2 * $29
key_down	fcb	2 * $21
key_up		fcb	2 * $31
key_Home	fcb	2 * $11 ; Skip to start of game
key_E		fcb	2 * $1d	; Skip to end of game
key_U		fcb	2 * $04	; Undo
key_R		fcb	2 * $15	; Redo
key_space	fcb	2 * $20	; Play
key_P		fcb	2 * $1c	; Pass

key_processor
		fcb	process_arrow - key_bsr
		fcb	process_arrow - key_bsr
		fcb	process_arrow - key_bsr
		fcb	process_arrow - key_bsr
		fcb	process_key_Home - key_bsr
		fcb	process_key_E - key_bsr
		fcb	process_key_U - key_bsr
		fcb	process_key_R - key_bsr
		fcb	process_key_space - key_bsr
		fcb	process_key_P - key_bsr

relative_dir	fcb	+(height + 2) - 1
		fcb	-1 - (height + 2)
		fcb	-(height + 2) + 1
direction	fcb	+(height + 2)
		fcb	-(height + 2)
		fcb	+1
		fcb	-1

char_table
chr_empty
		fcb	%00000000
		fcb	%00000000
		fcb	%00000000
		fcb	%00001000
		fcb	%00000000
		fcb	%00000000
		fcb	%00000000
		fcb	%00000000
chr_black
		fcb	%00111110
		fcb	%01111111
		fcb	%01111111
		fcb	%01111111
		fcb	%01111111
		fcb	%01111111
		fcb	%00111110
		fcb	%00000000
chr_white
		fcb	%00111110
		fcb	%01000001
		fcb	%01000001
		fcb	%01000001
		fcb	%01000001
		fcb	%01000001
		fcb	%00111110
		fcb	%00000000
chr_outside
		fcb	%00000000
		fcb	%00000000
		fcb	%00000000
		fcb	%00000000
		fcb	%00000000
		fcb	%00000000

wanted_page
data_size	equ	end_of_direct_page - direct_page

		if (wanted_page >> 8) != ((wanted_page + data_size - 1) >> 8)
		fill	$00, (((wanted_page >> 8) + 1) << 8) - wanted_page
		endif

direct_page
zeroes
		fcb	%00000000
		fcb	%00000000

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Initialized variables
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
key		fcb	key_Home
game_end	fdb	history
cursor		fdb	board + (1 + (width >> 1)) + (width + 2) * (1 + (height >> 1))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Uninitialized variables
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
to_move		rmb	1
opponent	rmb	1
ko		rmb	2
cursor_location	rmb	2

move_address	rmb	2
string_stone	rmb	1
liberties	rmb	1
string_size	rmb	1
capture_size	rmb	1
loop_index_1	rmb	1
loop_index_2
system_bottom	rmb	2

game_index	rmb	2

end_of_direct_page

board		rmb	(height + 2) * (width + 2)
max_game_length	equ	width * height
history		rmb	2 * max_game_length
stack_size	equ	2 * width * height
stack		rmb	stack_size
stack_bottom
